# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('perimeter', '0002_auto_20150611_1234'),
    ]

    operations = [
        migrations.AlterField(
            model_name='accesstokenuse',
            name='user_email',
            field=models.EmailField(max_length=254, verbose_name=b'Token used by (email)'),
        ),
    ]



# -*- coding: utf-8 -*-
"""Django models for the Perimeter app."""
from django.contrib.admin import site, ModelAdmin

from perimeter.models import AccessToken, AccessTokenUse


class AccessTokenAdmin(ModelAdmin):

    raw_id_fields = ('created_by',)
    list_display = ('token', 'expires_on', 'is_active', 'created_at', 'created_by')

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        """
        Presets the 'created_by' field to the current user.

        Taken from http://stackoverflow.com/a/5633217
        """
        if db_field.name == 'created_by':
            kwargs['initial'] = request.user.id
        return super(AccessTokenAdmin, self).formfield_for_foreignkey(
            db_field, request, **kwargs
        )


class AccessTokenUseAdmin(ModelAdmin):
    list_display = ('token', 'user_name', 'timestamp', 'client_ip')


site.register(AccessToken, AccessTokenAdmin)
site.register(AccessTokenUse, AccessTokenUseAdmin)



# -*- coding: utf-8 -*-
"""
Middleware component of Perimeter app - checks all incoming requests for a
valid token. See Perimeter docs for more details.
"""
import urllib

from django.core.exceptions import MiddlewareNotUsed
from django.core.urlresolvers import reverse
from django.http import HttpResponseRedirect
try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:
    # Fallback for Django < 1.10
    MiddlewareMixin = object

from perimeter.models import AccessToken
from perimeter.settings import (
    PERIMETER_SESSION_KEY,
    PERIMETER_ENABLED,
    PERIMETER_BYPASS_FUNCTION as bypass_perimeter
)


def get_request_token(request):
    """Returns AccessToken if found else EmptyToken."""
    assert hasattr(request, 'session'), (
        "Missing session attribute - please check MIDDLEWARE_CLASSES for "
        "'django.contrib.sessions.middleware.SessionMiddleware'."
    )
    token_value = request.session.get(PERIMETER_SESSION_KEY, None)
    # NB this method implements caching, so is more performant
    # than the straight get() alternative
    return AccessToken.objects.get_access_token(token_value)


def set_request_token(request, token_value):
    """Sets the request.session token value.

    Args:
        token - string, the token value (not the token object, as that is
            not serializable)
    """
    assert hasattr(request, 'session'), (
        "Missing session attribute - please check MIDDLEWARE_CLASSES for "
        "'django.contrib.sessions.middleware.SessionMiddleware'."
    )
    request.session[PERIMETER_SESSION_KEY] = token_value


class PerimeterAccessMiddleware(MiddlewareMixin):
    """
    Middleware used to detect whether user can access site or not.

    This middleware will be disabled if the PERIMETER_ENABLED setting does not
    exist in django settings, or is False.
    """
    def __init__(self, *args, **kwargs):
        """
        Disable middleware if PERIMETER_ENABLED setting not True.

        Raises MiddlewareNotUsed exception if the PERIMETER_ENABLED setting
        is not True - this is used by Django framework to remove the middleware.
        """
        if PERIMETER_ENABLED is False:
            raise MiddlewareNotUsed()
        super(PerimeterAccessMiddleware, self).__init__(*args, **kwargs)

    def process_request(self, request):
        """Check user session for token."""
        if bypass_perimeter(request):
            return None

        if get_request_token(request).is_valid:
            return None

        # redirect to the gateway for validation,
        qstring = urllib.urlencode({'next': request.get_full_path()})
        return HttpResponseRedirect(
            reverse('perimeter:gateway') + '?' + qstring
        )



# -*- coding: utf-8 -*-
# urls for perimeter app.
from django.conf.urls import url

from perimeter import views

urlpatterns = [
    url(
        r'^gateway/',
        views.gateway,
        name='gateway'
    ),
]



# -*- coding: utf-8 -*-
# Perimeter app views
import urllib

from django.core.urlresolvers import reverse, resolve, Resolver404
from django.http import HttpResponseRedirect
from django.shortcuts import render

from perimeter.forms import GatewayForm


def resolve_return_url(return_url):
    """Resolve a URL to confirm that it's valid.

    Before redirecting to a return_url, use the resolve function
    to confirm that it matches a valid view function.

    If the resolver can't find the function, return the only url we
    know exists - perimeter:gateway

    """
    path = None
    if return_url:
        return_url = urllib.unquote(return_url)
        # Path with a query string will not resolve, so we strip it for the check.
        path = return_url.split("?")[0]

    try:
        resolve(path)
        return return_url
    except Resolver404:
        return reverse('perimeter:gateway')


def gateway(request, template_name='perimeter/gateway.html'):
    """Display gateway form and process access requests.

    When the PerimeterAccessMiddleware catches an unvalidated
    user request they will redirect to this page.

    """
    if request.method == 'GET':
        form = GatewayForm()

    elif request.method == 'POST':
        form = GatewayForm(request.POST)
        if form.is_valid():
            form.save(request)
            return HttpResponseRedirect(
                resolve_return_url(
                    request.GET.get('next'))
            )

    return render(request, template_name, {'form': form})



# -*- coding: utf-8 -*-
# perimeter form tests
import datetime

from django.core.exceptions import ValidationError
from django.test import TestCase, RequestFactory, override_settings
from django.utils.timezone import now

from perimeter.forms import GatewayForm
from perimeter.models import AccessToken, AccessTokenUse

YESTERDAY = now().date() - datetime.timedelta(days=1)

class GatewayFormTests(TestCase):

    def setUp(self):
        self.factory = RequestFactory()
        self.payload = {
            'token': "test",
            'email': "hugo@yunojuno.com",
            'name': "Hugo Rodger-Brown"
        }
        self.request = self.factory.post('/',
            data=self.payload,
            REMOTE_ADDR='127.0.0.1',
            HTTP_USER_AGENT='test_agent'
        )
        self.token = AccessToken(token="test").save()
        self.form = GatewayForm(self.request.POST)

    def test_post_valid_token(self):
        self.assertTrue(self.form.is_valid())
        self.assertEqual(self.form.token, self.token)

    def test_clean_inactive_token(self):
        self.token.is_active = False
        self.token.save(update_fields=['is_active'])
        self.assertFalse(self.form.is_valid())
        self.assertRaises(ValidationError, self.form.clean_token)

    def test_clean_expired_token(self):
        self.token.expires_on = YESTERDAY
        self.token.save(update_fields=['expires_on'])
        self.assertFalse(self.form.is_valid())
        self.assertRaises(ValidationError, self.form.clean_token)

    def test_no_matching_token(self):
        AccessToken.objects.all().delete()
        self.assertFalse(self.form.is_valid())
        self.assertRaises(ValidationError, self.form.clean_token)

    def test_save(self):
        self.request.session = {}
        self.assertTrue(self.form.is_valid())
        au = self.form.save(self.request)
        self.assertTrue(AccessTokenUse.objects.get(), au)
        self.assertEqual(au.user_email, self.payload['email'])
        self.assertEqual(au.user_name, self.payload['name'])
        self.assertEqual(au.token, self.token)
        self.assertEqual(au.client_ip, '127.0.0.1')
        self.assertEqual(au.client_user_agent, 'test_agent')



# -*- coding: utf-8 -*-
"""
Django models for the Perimeter app.
"""
from datetime import date, time, timedelta, datetime
import random

from django.conf import settings
from django.core.cache import cache
from django.db import models
from django.db.models.signals import post_save, post_delete
from django.dispatch import receiver
from django.utils.timezone import now, make_aware, is_naive, is_aware, get_current_timezone

from perimeter.settings import PERIMETER_DEFAULT_EXPIRY


def default_expiry():
    """Return the default expiry date."""
    return (now() + timedelta(days=PERIMETER_DEFAULT_EXPIRY)).date()


class EmptyToken(object):
    """Token-like object that will always return is_valid() == False.

    EmptyToken objects are a bit like Django's AnonymousUser model -
    they return an object that can be used like an AccessToken but that
    is always invalid.
    """
    @property
    def is_valid(self):
        return False


class AccessTokenManager(models.Manager):
    """Custom model manager for AccessTokens."""

    def create_access_token(self, **kwargs):
        """Create a new AccessToken with a random token value."""
        # NB there is a theoretical token clash exception here,
        # when the random_token_value function returns an existing
        # token value, but it is considered so unlikely as to be
        # acceptable.
        kwargs['token'] = kwargs.get('token', AccessToken.random_token_value())
        kwargs['expires_on'] = kwargs.get('expires_on', default_expiry())
        return AccessToken(**kwargs).save()

    def get_access_token(self, token_value):
        """Fetch an AccessToken, return EmptyToken if not found.

        This method is cache-aware, and will check the cache first,
        re-filling it if empty.

        Args:
            token - string, the token value to look up.

        """
        cache_key = AccessToken.get_cache_key(token_value)
        token = cache.get(cache_key)
        if token is None:
            try:
                token = self.get(token=token_value)
                cache.set(token.cache_key, token, token.seconds_to_expiry)
                return token
            except AccessToken.DoesNotExist:
                return EmptyToken()
        else:
            return token


class AccessToken(models.Model):
    """
    A token that allows a user entry to the site via Perimeter.
    """
    token = models.CharField(max_length=50, unique=True)
    is_active = models.BooleanField(default=True)
    # NB pass in a callable, not the result of the callable, see:
    # http://stackoverflow.com/a/29549675/45698
    expires_on = models.DateField(default=default_expiry)
    created_by = models.ForeignKey(settings.AUTH_USER_MODEL, null=True)
    created_at = models.DateTimeField()
    updated_at = models.DateTimeField()

    objects = AccessTokenManager()

    def __unicode__(self):
        if self.is_valid:
            return (
                u"%s - valid until %s" % (
                    self.token,
                    self.expires_on
                )
            )
        elif self.is_active is False:
            return (
                u"%s - inactive" % self.token
            )
        elif self.has_expired:
            return (
                u"%s - expired on %s" % (
                    self.token,
                    self.expires_on
                )
            )
        else:
            return (u"%s - invalid" % self.token)

    def __str__(self):
        return self.__unicode__().encode('UTF-8')

    @classmethod
    def random_token_value(cls):
        """Generate a random token value."""
        return "".join(
            random.sample(
                population=list(
                    "abcdefghijklmnopqrstuvwxyz"
                    "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                    "1234567890"
                ),
                k=cls._meta.get_field('token').max_length
            )
        )

    @classmethod
    def get_cache_key(cls, token_value):
        return (
            '%s.%s-%s' % (
                cls.__module__,
                cls.__name__,
                token_value
            )
        )

    def save(self, *args, **kwargs):
        "Sets the created_at timestamp."
        self.updated_at = now()
        self.created_at = self.created_at or self.updated_at
        super(AccessToken, self).save(*args, **kwargs)
        return self

    @property
    def cache_key(self):
        """Return object cache key (from get `get_cache_key`)."""
        return AccessToken.get_cache_key(self.token)

    @property
    def seconds_to_expiry(self):
        """Return the number of seconds till expiry (used for caching)."""
        expires_at = datetime.combine(self.expires_on, time.min)
        if settings.USE_TZ and is_naive(expires_at):
            expires_at = make_aware(expires_at, get_current_timezone())
        elif not settings.USE_TZ and is_aware(expires_at):
            expires_at = make_naive(expires_at)
        return int((expires_at - now()).total_seconds())

    @property
    def has_expired(self):
        """Return True if the token has passed expiry date."""
        return self.expires_on < date.today()

    @property
    def is_valid(self):
        """Return True if the token is active and has not expired."""
        return self.is_active and not self.has_expired

    def record(
        self,
        user_email,
        user_name,
        client_ip='unknown',
        client_user_agent='unknown'
    ):
        """Record the fact that someone has used the token."""
        atu = AccessTokenUse(
            token=self,
            user_email=user_email,
            user_name=user_name,
            client_ip=client_ip,
            client_user_agent=client_user_agent
        )
        atu.save()
        return atu


@receiver(post_save, sender=AccessToken)
def on_save_access_token(sender, instance, **kwargs):
    """Update saved object in cache if is_valid, else delete."""
    cache.set(instance.cache_key, instance, instance.seconds_to_expiry)


@receiver(post_delete, sender=AccessToken)
def on_delete_access_token(sender, instance, **kwargs):
    """Remove deleted object from cache."""
    cache.delete(instance.cache_key)


class AccessTokenUse(models.Model):
    """Audit record used to log whenever an access token is used."""
    token = models.ForeignKey(AccessToken)
    user_email = models.EmailField(
        verbose_name="Token used by (email)")
    user_name = models.CharField(max_length=100,
        verbose_name="Token used by (name)")
    client_ip = models.CharField(
        max_length=15,
        verbose_name='IP address',
        blank=True
    )
    client_user_agent = models.CharField(
        max_length=150,
        verbose_name="User Agent",
        blank=True
    )
    timestamp = models.DateTimeField()

    def __unicode__(self):
        return u"'%s' used by '%s'" % (self.token.token, self.user_email)

    def __str__(self):
        return self.__unicode__().encode('UTF-8')

    def save(self, *args, **kwargs):
        "Set the timestamp and save the object."
        self.timestamp = self.timestamp or now()
        super(AccessTokenUse, self).save(*args, **kwargs)
        return self
